@@ -7,13 +7,13 @@
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program (see the file COPYING); if not, write to the
  * Free Software Foundation, Inc.,
- * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  *
  */
 
@@ -29,32 +29,30 @@
 #include <fcntl.h>
 #include <time.h>
 #include <errno.h>
+#include <signal.h>
 #include <sys/ioctl.h>
 #include <sys/mman.h>
 #include <sys/types.h>
 #include <sys/wait.h>
-#include <linux/videodev.h>
+#include <libv4l1-videodev.h>
 #include <linux/soundcard.h>
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/Xmd.h>
 #include <X11/StringDefs.h>
-#include <X11/keysym.h>
+#include <X11/XKBlib.h>
 #include <X11/xpm.h>
 #include <X11/extensions/shape.h>
-#include <X11/extensions/xf86dga.h>
+#include <X11/extensions/Xxf86dga.h>
 #include <X11/extensions/xf86vmode.h>
 
 #include "wmgeneral/wmgeneral.h"
-#include "wmgeneral/misc.h"
 #include "wmtv-master.xpm"
 #include "channels.h"
 
 
 /* Defines */
-#define VERSION "0.6.5"
-
 #define ON		1
 #define OFF		0
 
@@ -75,15 +73,19 @@
 #define SETUNE		2
 #define SETSPD		3
 
-#define MAXCHAN		100
-#define OPTIONS		"hvd:g:e:b:"
+#define MAXCHAN		99
+#define OPTIONS		"hvd:g:e:b:c:"
 
 #define TELEVISION	0
 #define COMPOSITE   1
 #define SVIDEO		2
 
+#ifndef GLOBALCONFFILE
+#define GLOBALCONFFILE "/etc/wmtvrc"
+#endif
 
 /* Global Variables */
+Display		*display;
 int tfd;
 int card_present = FALSE;
 int ntfb_status  = SETOFF;
@@ -111,11 +113,12 @@ int dcret;
 int tml;
 int fswidth = 0;
 int fsheight = 0;
+pid_t child_pid = -1;
+int restart = FALSE;
 
 unsigned long ccrfreq;
 unsigned long rfreq;
 unsigned long st;
-unsigned long offset;
 
 char *norm = NULL;
 char *source;
@@ -123,11 +126,14 @@ char *mode = NULL;
 char *fullscreen = NULL;
 int  freqnorm = -1;
 char *cname[MAXCHAN];
-char *wcname[MAXCHAN];
-int	 ftune[MAXCHAN];
+char *comment[MAXCHAN];
+long ftune[MAXCHAN];
 char *progname;
 char *dev = "/dev/video";
 
+char *sysConfFile = GLOBALCONFFILE;
+char *usrConfFile;
+
 int  wmtv_mask_width = 64;
 int  wmtv_mask_height = 64;
 char wmtv_mask_bits[64*64];
@@ -196,22 +202,91 @@ void GrabImage(void);
 
 void ParseRCFile(const char *, rckeys *);
 void ParseRCFile2(const char *);
-void WriteRCFile(const char *);
+void WriteRCFile(const char *, rckeys *);
 void InitConfig(void);
 void InitPalette(void);
 
 void Usage(void);
 void Version(void);
 
+void
+sigchld_handler(int i)
+{
+	pid_t pid;
+	pid = waitpid((pid_t)-1, NULL, WNOHANG);
+	while (pid>(pid_t)0) {
+		if (pid == child_pid) {
+			child_pid = -1;
+			restart = TRUE;
+		}
+		pid = waitpid((pid_t)-1, NULL, WNOHANG);
+	}
+}
+
+void
+init_signal()
+{
+	struct sigaction sa;
+
+	sa.sa_handler = &sigchld_handler;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_RESTART;
+	sigaction(SIGCHLD, &sa, NULL);
+}
+
+char *
+expand_format(char *format, char *letters, char **expansions)
+{
+	char *string;
+	unsigned int string_pos = 0;
+	unsigned int size;
+	size = strlen(format)+1;
+	string = (char *)malloc(size*sizeof(char));
+	for (; *format != '\0'; format++) {
+		if (*format == '%') {
+			format++;
+			if (*format != '%') {
+				unsigned int i;
+				for (i = 0; letters[i] != '\0'; i++) {
+					if (letters[i] == *format)
+						break;
+				}
+				if (letters[i] != '\0') {
+					unsigned int expansion_size;
+					expansion_size = strlen(expansions[i]);
+					while (string_pos+expansion_size+1 > size) {
+						size *= 2;
+						string = realloc(string, size*sizeof(char));
+					}
+					memcpy(string+string_pos, expansions[i], expansion_size);
+					string_pos += expansion_size;
+					continue;
+				}
+				else {
+					format--;
+				}
+			}
+		}
+		while (string_pos+1+1 > size) {
+			size *= 2;
+			string = realloc(string, size*sizeof(char));
+		}
+		string[string_pos] = *format;
+		string_pos++;
+	}
+	string[string_pos] = '\0';
+	return string;
+}
 
 /* main function */
 int
 main(int argc, char *argv[])
 {
-	int i, c, opind;
-	// pid_t pid;
-	char cfile[128];
+	int c, opind;
+	int pressed_button = -1;
+	/* pid_t pid; */
 	static struct option long_options[] = {
+		{"device", 1, 0, 'c'},
 		{"display", 1, 0, 'd'},
 		{"geometry", 1, 0, 'g'},
 		{"bpp", 1, 0, 'b'},
@@ -223,8 +298,18 @@ main(int argc, char *argv[])
 
 	progname = strdup(argv[0]);
 
-	strncpy(cfile, (char *)getenv("HOME"), sizeof(char)*128);
-	strcat(cfile, "/.wmtvrc");
+	{
+		char *home = getenv("HOME");
+
+		if (home == NULL) {
+			fprintf(stderr, "wmtv: $HOME should be set.\n");
+			exit(1);
+		}
+
+		usrConfFile = (char *)malloc(sizeof(char)*(strlen(home)+8+1));
+		strcpy(usrConfFile, home);
+		strcat(usrConfFile, "/.wmtvrc");
+	}
 
 	while (1) {
 		opind = 0;
@@ -232,7 +317,7 @@ main(int argc, char *argv[])
 		if (c == -1)
 		break;
 
-		switch(c) 
+		switch(c)
 			{
 				case 0:
 					/*
@@ -250,7 +335,10 @@ main(int argc, char *argv[])
 					break;
 				case 'e':
 					exe = strdup(optarg);
-					strcat(exe, " &");
+					/* strcat(exe, " &"); */
+					break;
+				case 'c':
+					dev = strdup(optarg);
 					break;
 				case 'b':
 					fprintf(stderr, "wmtv: option not implemented yet\n");
@@ -268,19 +356,31 @@ main(int argc, char *argv[])
 	}
 
 	/* creat windows */
-	createXBMfromXPM (wmtv_mask_bits, wmtv_master_xpm, 
+	createXBMfromXPM (wmtv_mask_bits, wmtv_master_xpm,
 					wmtv_mask_width, wmtv_mask_height);
 	openXwindow (argc, argv, wmtv_master_xpm, wmtv_mask_bits,
-					wmtv_mask_width, wmtv_mask_height); 
+					wmtv_mask_width, wmtv_mask_height);
 
 	AddMouseRegion (NTFB, 47, 48, 59, 59);		/* On/SetTune/Off Button */
 	AddMouseRegion (SCANLB, 23, 48, 35, 59);    /* Left Preset/Scan Button */
 	AddMouseRegion (SCANRB, 35, 48, 47, 59); 	/* Right Preset/Scan Button */
 	AddMouseRegion (FULLSB, 5, 5, 59, 44);		/* Toggle FullScreen */
 
+	init_signal();
+
 	/* wmtv main loop */
-	while (1) 
+	while (1)
 	{
+		if (restart) {
+			TVOn();
+			if (ntfb_status == SETOFF) {
+				ntfb_status = SETON;
+				RedrawWindow();
+				XFlush(display);
+			}
+			restart = FALSE;
+			continue;
+		}
 		while (XPending(display))
 		{
 			XNextEvent(display, &Event);
@@ -303,10 +403,10 @@ main(int argc, char *argv[])
 						RetScreen();
 					}
 					else {
-					i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
-					switch (i) {
+					pressed_button = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
+					switch (pressed_button) {
 						case NTFB:
-							ButtonDown(NTFB);	
+							ButtonDown(NTFB);
 							t_lc = Event.xbutton.time;
 							but_pressed = TRUE;
 							break;
@@ -314,7 +414,7 @@ main(int argc, char *argv[])
 							ButtonDown(SCANLB);
 							if (ntfb_status == SETUNE) {
 								switch (Event.xbutton.button) {
-									case 1: 
+									case 1:
 										timebutton = 1;
 										while (timebutton == 1)
 										if (isource == TELEVISION)
@@ -326,10 +426,10 @@ main(int argc, char *argv[])
 										break;
 								}
 							}
-							else 
+							else if (ntfb_status == SETON)
 									ChanDown();
 							break;
-						case SCANRB: 
+						case SCANRB:
 							ButtonDown(SCANRB);
 							if (ntfb_status == SETUNE) {
 								switch (Event.xbutton.button) {
@@ -345,7 +445,7 @@ main(int argc, char *argv[])
 										break;
 								}
 							}
-							else 
+							else if (ntfb_status == SETON)
 									ChanUp();
 							break;
 						case FULLSB:
@@ -372,9 +472,8 @@ main(int argc, char *argv[])
 					}
 					break;
 				case ButtonRelease:
-					i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
-					switch (i) {
-						case NTFB: 
+					switch (pressed_button) {
+						case NTFB:
 							ButtonUp(NTFB);
 								if (but_pressed) {
 									if (Event.xbutton.time - t_lc >= 900) {
@@ -384,13 +483,15 @@ main(int argc, char *argv[])
 								}
 
 								if (ntfb_status == SETOFF) {
+									if (child_pid == -1) {
 										ntfb_status = SETON;
 										TVOn();
 								}
+								}
 								else if (ntfb_status == SETON) {
 										if (!btime) {
 										ntfb_status = SETUNE;
-										copyXPMArea(96, 79, 11, 7, 6, 50); 
+										copyXPMArea(96, 79, 11, 7, 6, 50);
 										RedrawWindowXYWH(6, 50, 11, 7);
 										}
 										else if (btime) {
@@ -402,9 +503,9 @@ main(int argc, char *argv[])
 								}
 								else if (ntfb_status == SETUNE) {
 										if (!btime) {
-											offset = (rfreq - ccrfreq);
-											// fprintf(stderr, "wmtv: finetune offset = %ld\n", offset);
-											WriteRCFile(cfile); 
+											ftune[cchannel] = (rfreq - ccrfreq);
+											/* fprintf(stderr, "wmtv: finetune offset = %ld\n", ftune[cchannel]); */
+											WriteRCFile(usrConfFile, wmtv_keys);
 											ntfb_status = SETON;
 											DrawPresetChan(cchannel);
 										}
@@ -438,13 +539,43 @@ main(int argc, char *argv[])
 											{
 												if ((ntfb_status == SETON) || (ntfb_status == SETUNE)) {
 													if (exe) {
+														pid_t pid;
+														char *command;
+														char *letters = "#nf";
+														char *(expansions[3]);
 														ntfb_status = SETOFF;
 														TVOff();
-														system(exe);
-														/*
+														expansions[0] = malloc(3*sizeof(char));
+														snprintf(expansions[0], 3, "%d", cchannel+1);
+														expansions[1] = comment[cchannel];
+														expansions[2] = malloc(15*sizeof(char));
+														snprintf(expansions[2], 15, "%.3f", (double)rfreq/(double)st);
+														command = expand_format(exe, letters, expansions);
+														/* system(exe); */
+														child_pid = fork();
+														if (child_pid == (pid_t) 0) {
+														char *argv[4];
+														argv[0] = "sh";
+														argv[1] = "-c";
+															argv[2] = command;
+														argv[3] = NULL;
+														execv("/bin/sh", argv);
+														exit(-1);
+														}
+														free(expansions[0]);
+														free(expansions[2]);
+														free(command);
+														pid = waitpid(child_pid, NULL, WNOHANG);
+														if (pid != 0) {
+															child_pid = -1;
+															restart = TRUE;
+														}
+														/* printf("Returned pid:\n"); */
+
+#if 0
 														pid = fork();
 
-														// child
+														/* child */
 														if (pid == (pid_t) 0) {
 														execlp("xawtv", "xawtv", "&", (char *) 0);
 														}
@@ -453,11 +584,11 @@ main(int argc, char *argv[])
 															perror("fork");
 														}
 
-														// parent
+														/* parent */
 														else {
 															if (waitpid(pid, NULL, 0) < 0) {
 																perror("waitpid");
-														*/
+#endif
 													}
 													else {
 														DoFullScreen();
@@ -535,7 +666,7 @@ main(int argc, char *argv[])
 					break;
 				case KeyPress:
 					if ((Event.xany.window == fmwin) && fmode) {
-							switch(XKeycodeToKeysym(display, Event.xkey.keycode, 0)) {
+							switch(XkbKeycodeToKeysym(display, Event.xkey.keycode, 0, 0)) {
 								case XK_Up:
 									if (isource == TELEVISION)
 									ChanUp();
@@ -545,12 +676,10 @@ main(int argc, char *argv[])
 									ChanDown();
 									break;
 								case XK_Right:
-									if (isource == TELEVISION)
-									FineTuneUp();
+									VolumeUp();
 									break;
 								case XK_Left:
-									if (isource == TELEVISION)
-									FineTuneDown();
+									VolumeDown();
 									break;
 								case XK_m:
 									if (!mute) {
@@ -569,13 +698,13 @@ main(int argc, char *argv[])
 					}
 				default:
 					break;
-			} 
+			}
 		XFlush(display);
-		}  
+		}
 	    usleep(50000L);
-	} 
+	}
 	return(0);
-} 
+}
 
 
 /* ButtonDown function */
@@ -633,7 +762,6 @@ TVOn(void)
 	int rx, ry;
 	char *p;
 
-
 	XWindowAttributes winattr;
 
 	if (!XGetWindowAttributes (display, iconwin, &winattr)) {
@@ -647,6 +775,18 @@ TVOn(void)
 		fprintf(stderr, "wmtv: error translating coordinates\n");
 	}
 
+	/* This part was taken from xawtv's source code */
+	switch (system("v4l-conf -q")) {
+	case -1: /* can't run */
+		fprintf(stderr,"could'nt start v4l-conf\n");
+		break;
+	case 0: /* ok */
+		break;
+	default: /* non-zero return */
+		fprintf(stderr,"v4l-conf had some trouble, "
+				"trying to continue anyway\n");
+	}
+	/* End of "stolen" part */
 	InitConfig();
 	GetFrameBuffer();
 	InitPalette();
@@ -669,7 +809,7 @@ TVOn(void)
 	vclip[1].height = 39;
 
 	vwin.clips  = vclip;
-	vwin.clipcount = 2; 
+	vwin.clipcount = 2;
 
 	vchn.channel = tvsource;
 	vaud.audio = tvsource;
@@ -691,9 +831,9 @@ TVOn(void)
 	if (!cnotune) {
 		if (ioctl(tfd, VIDIOCGTUNER, &vtun) < 0)
 			perror("ioctl VIDIOCGTUNER");
-		if (vtun.flags & VIDEO_TUNER_LOW) 
+		if (vtun.flags & VIDEO_TUNER_LOW)
 			st = 16000;
-		else 
+		else
 			st = 16;
 	}
 
@@ -707,7 +847,7 @@ TVOn(void)
 
 		if (vaud.flags & VIDEO_AUDIO_MUTE) {
 			vaud.flags &= ~VIDEO_AUDIO_MUTE; 	/* Unmute */
-			vaud.volume = (0xFFFF/2)+1;
+			vaud.volume = 0xFFFF;
 		}
 
 	if (isource == TELEVISION) {
@@ -743,7 +883,7 @@ TVOn(void)
 
 	if (ioctl(tfd, VIDIOCSWIN, &vwin) < 0)
 			perror("ioctl VIDIOCSWIN");
-	
+
 	if (ioctl(tfd, VIDIOCCAPTURE, &ccapt) < 0)
 			perror("ioctl VIDIOCCAPTURE");
 }
@@ -755,7 +895,7 @@ TVOff(void)
 {
 	ccapt = 0;
 
-	if (ioctl(tfd, VIDIOCCAPTURE, &ccapt) < 0) 
+	if (ioctl(tfd, VIDIOCCAPTURE, &ccapt) < 0)
 			perror("ioctl VIDIOCCAPTURE");
 
 	vaud.audio  = tvsource;
@@ -763,7 +903,7 @@ TVOff(void)
 	vaud.flags |= VIDEO_AUDIO_MUTE;
 
 	if (vchn.flags & VIDEO_VC_AUDIO) {
-		if (ioctl(tfd, VIDIOCSAUDIO, &vaud) < 0) 
+		if (ioctl(tfd, VIDIOCSAUDIO, &vaud) < 0)
 				perror("ioctl VIDIOCSAUDIO");
 	}
 
@@ -778,14 +918,16 @@ void
 VolumeUp(void)
 {
 	if(vchn.flags & VIDEO_VC_AUDIO) {
-		if (vaud.volume <= 0xFFFF) {
+		if ((vaud.flags & VIDEO_AUDIO_VOLUME) == 0)
+			fprintf(stderr, "Warning: v4l device does not support volume control.\n");
+		else if (vaud.volume <= 0xEEEE) {
 			vaud.audio = tvsource;
-			vaud.volume += (0xFFFF/10);
+			vaud.volume += 0x1111;
 			if (ioctl(tfd, VIDIOCSAUDIO, &vaud) < 0)
 				perror("ioctl VIDIOCSAUDIO");
 		}
 	}
-		
+
 }
 
 /* VoumeDown function */
@@ -793,9 +935,11 @@ void
 VolumeDown(void)
 {
 	if (vchn.flags & VIDEO_VC_AUDIO) {
-		if (vaud.volume > 0) {
+		if ((vaud.flags & VIDEO_AUDIO_VOLUME) == 0)
+			fprintf(stderr, "Warning: v4l device does not support volume control.\n");
+		else if (vaud.volume >= 0x1111) {
 			vaud.audio = tvsource;
-			vaud.volume -= (0xFFFF/10);
+			vaud.volume -= 0x1111;
 			if (ioctl(tfd, VIDIOCSAUDIO, &vaud) < 0)
 				perror("ioctl VIDIOCSAUDIO");
 		}
@@ -808,7 +952,7 @@ MuteAudio(void)
 {
 	if (vchn.flags & VIDEO_VC_AUDIO) {
 		vaud.audio = tvsource;
-		// vaud.volume = 0;
+		/* vaud.volume = 0; */
 		vaud.flags |= VIDEO_AUDIO_MUTE;
 		if (ioctl(tfd, VIDIOCSAUDIO, &vaud) < 0)
 				perror("ioctl VIDIOCSAUDIO");
@@ -823,7 +967,7 @@ UnMuteAudio(void)
 {
 	if ((vchn.flags & VIDEO_VC_AUDIO) && (vaud.flags & VIDEO_AUDIO_MUTE)) {
 		vaud.audio = tvsource;
-		// vaud.volume = (0xFFFF/2)+1;
+		/* vaud.volume = (0xFFFF/2)+1; */
 		vaud.flags &= ~VIDEO_AUDIO_MUTE;
 		if (ioctl(tfd, VIDIOCSAUDIO, &vaud) < 0)
 				perror("ioctl VIDIOCSAUDIO");
@@ -833,7 +977,7 @@ UnMuteAudio(void)
 
 
 /* ScanUp function */
-void 
+void
 ScanUp(void)
 {
 	rfreq += 2;
@@ -849,7 +993,7 @@ ScanUp(void)
 	if (vtun.signal == 0xFFFF) {
 		timebutton = 0;
 	}
-	
+
 }
 
 
@@ -900,14 +1044,14 @@ void
 ChanUp(void)
 {
 	int i;
-	if (cchannel != maxpst)
+	if (cchannel < maxpst)
 	++cchannel;
 	for (i=0; i < CHAN_ENTRIES; i++) {
 		if (!strcmp(cname[cchannel], tvtuner[i].name)) {
 				rfreq = ((tvtuner[i].freq[freqnorm] / 1000) * st) + ftune[cchannel];
 				ccrfreq = rfreq - ftune[cchannel];
 		break;
-		}	
+		}
 	}
 
 	DrawPresetChan(cchannel);
@@ -917,12 +1061,12 @@ ChanUp(void)
 
 
 /* ChanDown function */
-void 
+void
 ChanDown(void)
 {
 	int i;
-	if (cchannel != 0) 
-	--cchannel;			
+	if (cchannel != 0)
+	--cchannel;
 	for (i=0; i < CHAN_ENTRIES; i++) {
 		if (!strcmp(cname[cchannel], tvtuner[i].name)) {
 				rfreq = ((tvtuner[i].freq[freqnorm] / 1000) * st) + ftune[cchannel];
@@ -947,7 +1091,7 @@ DrawPresetChan(int cchannel)
 	int k=6;
 
 	if (isource == TELEVISION) {
-		sprintf(temp, "%02d", cchannel);
+		sprintf(temp, "%02d", cchannel+1);
 
 		if (*p == '0') {
 			copyXPMArea(66, 79, 5, 7, k, 50);
@@ -977,10 +1121,10 @@ DrawPresetChan(int cchannel)
 
 
 /* ParseRCFile function */
-void 
-ParseRCFile(const char *filename, rckeys *keys) 
+void
+ParseRCFile(const char *filename, rckeys *keys)
 {
-	char	*p,*q;
+	char	*p;
 	char	temp[128];
 	char	*tokens = " =\t\n";
 	FILE	*fp;
@@ -993,36 +1137,30 @@ ParseRCFile(const char *filename, rckeys *keys)
 	}
 	norcfile = 0;
 	while (fgets(temp, 128, fp)) {
-		key = 0;
-		q = strdup(temp);
-		if (*q != '\n') {
-		q = strtok(q, tokens);
-			while (key >= 0 && keys[key].label) {
-				if ((!strcmp(q, keys[key].label))) {
+		if (temp[0] != '\n') {
+			p = strtok(temp, tokens);
+			for (key=0; keys[key].label; key++) {
+				if ((!strcmp(p, keys[key].label))) {
 					p = strtok(NULL, tokens);
 					free(*keys[key].var);
 					*keys[key].var = strdup(p);
-					key = -1;
-				} else key++;
-			} 
+					break;
+				}
+			}
 		}
-		free(q);
-	} 
+	}
 	fclose(fp);
 }
 
 
 /* ParseRCFile2 function */
 void
-ParseRCFile2(const char *filename) 
+ParseRCFile2(const char *filename)
 {
 	int  menu = FALSE;
-	char temp[128];
-	char tp[10];
-	char *tokens = " \t\n()";
-	char *q, *p;
+	char temp[128], name[128];
 	FILE *fp;
-	int i = 0;
+	int len, i = 0;
 
 	if ((fp = fopen(filename, "r")) == NULL) {
 		fprintf(stderr, "wmtv: %s\n", strerror(errno));
@@ -1031,79 +1169,56 @@ ParseRCFile2(const char *filename)
 	}
 	norcfile = 0;
 	while (fgets(temp, 128, fp)) {
-		q = strdup(temp);
-		if (*q != '\n') {
-			q = strtok(q, tokens);
+		if (*temp != '\n') {
 			if (menu) {
-				cname[i] = (char *)malloc(sizeof(q));
-				strncpy(cname[i], q, sizeof(q));
-				p = q;
-				p = strtok(NULL, tokens);
-				if (p != NULL) {
-					if (*p == '-') {
-						p++;
-						ftune[i] = -1*atoi(p);
-					}
-					else if (*p == '+') {
-						p++;
-						ftune[i] = atoi(p);
-					}
-					else {
-						ftune[i] = atoi(p);
+				ftune[i]=0;
+				if(sscanf(temp, " %s %n(%ld) %n", name, &len, &ftune[i], &len)>=1) {
+					char *pos;
+					cname[i]=strdup(name);
+					comment[i]=temp[len]?strdup(temp+len):"";
+					/* Remove the end-of-line symbol */
+					for (pos = comment[i]; *pos != '\0'; pos++) {
+						if (*pos == '\n') {
+							*pos = '\0';
+							break;
+						}
 					}
 				}
-				else {
-					ftune[i] = 0;
-				}
-				wcname[i] = (char *)malloc(sizeof(cname[i])+sizeof(p));
-				wcname[i] = strdup(cname[i]);
-				sprintf(tp, " (%d) ", ftune[i]);	
-				strcat(wcname[i], tp);
-				i++;
-				tpst = i;
+				if(++i>=MAXCHAN)
+					break;
 			}
-			if ((q = strchr(q, '[')) != NULL) {
+			else if (strchr(temp, '[')) {
 				menu = TRUE;
 			}
 		}
 	}
+	tpst = i;
 	fclose(fp);
 }
 
 
 /* WriteRCFile function */
 void
-WriteRCFile(const char *filename)
+WriteRCFile(const char *filename, rckeys *keys)
 {
-	int i;
-	char temp[128];
-	char tp[10];
-	char *tokens = " \t\n()";
-	char *q;
+	long i;
 	FILE *fp;
+	int key;
 
-	if ((fp = fopen(filename, "r+")) == NULL) {
+	if ((fp = fopen(filename, "w")) == NULL) {
 		fprintf(stderr, "wmtv: %s\n", strerror(errno));
 		return;
 	}
 
+	for (key=0; keys[key].label; key++)
+		if (*keys[key].var)
+			fprintf(fp, "%s = %s\n", keys[key].label, *keys[key].var);
+
+	fprintf(fp, "\n[channel]\n");
+
+			for (i = 0; i <= maxpst; i++)
+		fprintf(fp, "%s (%ld)\t%s\n", cname[i], ftune[i], comment[i]);
 
-	while (fgets(temp, 128, fp)) {
-		q = strdup(temp);
-		if (*q != '\n') {
-			q = strtok(temp, tokens);
-			if ((q = strchr(q, '[')) != NULL) {
-				for (i = 0; i <= maxpst; i++) {
-					// fprintf(stderr, "offset is %ld\n", offset);
-					sprintf(tp, " (%ld) ", offset);
-					strtok(wcname[cchannel], tokens);
-					strcat(wcname[cchannel], tp);
-					fputs(wcname[i], fp);
-					fputs("\n", fp);
-				}
-			}
-		}
-	}
 	fclose(fp);
 }
 
@@ -1155,20 +1270,18 @@ void
 InitConfig(void)
 {
 	int i;
-	char temp[128];
 
-	strcpy(temp, "/etc/wmtvrc");
-	ParseRCFile(temp, wmtv_keys);
-	ParseRCFile2(temp);
-	
-	strncpy(temp, (char *)getenv("HOME"), (sizeof(char)*128));
-	strcat(temp, "/.wmtvrc");
-	ParseRCFile(temp, wmtv_keys);
-	ParseRCFile2(temp);
+	ParseRCFile(usrConfFile, wmtv_keys);
+	ParseRCFile2(usrConfFile);
 
 	if (norcfile) {
-		fprintf(stderr, "wmtv: error - config file not found\n");
-		exit(1);
+		ParseRCFile(sysConfFile, wmtv_keys);
+		ParseRCFile2(sysConfFile);
+
+		if (norcfile) {
+			fprintf(stderr, "wmtv: error - config file not found\n");
+			exit(1);
+		}
 	}
 
 	if (maxpreset != NULL)
@@ -1202,7 +1315,7 @@ InitConfig(void)
 	}
 
 	if (!(vcap.type & VID_TYPE_SCALES)) {
-		fprintf(stderr, "%s: video device does not support scalling\n", progname); 
+		fprintf(stderr, "%s: video device does not support scalling\n", progname);
 		exit(1);
 	}
 	if (!(vcap.type & VID_TYPE_CLIPPING)) {
@@ -1229,7 +1342,7 @@ InitConfig(void)
 					cnotune = 0;
 				else
 					cnotune = 1;
-			
+
 				vsource = 1;
 				tvsource = vchn.channel;
 
@@ -1274,7 +1387,7 @@ InitConfig(void)
 				}
 		}
 	}
-		
+
 	for (i=0; i < vcap.audios; i++) {
 		vaud.audio = i;
 		if (ioctl(tfd, VIDIOCGAUDIO, &vaud) < 0) {
@@ -1297,23 +1410,26 @@ InitConfig(void)
 void
 GetFrameBuffer(void)
 {
+#if 0
 	void *baseaddr = NULL;
 	int evbr, erbr, flr = 0;
 	int bankr, memr, depth;
 	int i, n;
 	int bytesperline, bitsperpixel;
 	XPixmapFormatValues *pf;
+#endif
 
 	if (!XGetWindowAttributes(display, DefaultRootWindow(display), &Winattr)) {
 		fprintf(stderr, "wmtv: error getting winattr of root\n");
 	}
 
+#if 0
 	depth = Winattr.depth;
 
 	if (XF86DGAQueryExtension(display, &evbr, &erbr)) {
 			XF86DGAQueryDirectVideo(display, XDefaultScreen(display), &flr);
 			if (flr & XF86DGADirectPresent) {
-					XF86DGAGetVideoLL(display, XDefaultScreen(display), 
+					XF86DGAGetVideoLL(display, XDefaultScreen(display),
 							(int *) &baseaddr, &bytesperline, &bankr, &memr);
 			}
 	}
@@ -1331,8 +1447,8 @@ GetFrameBuffer(void)
 			}
 	}
 
-	bitsperpixel = (depth+7) & 0xf8;				    /* Taken from */
-	bytesperline = Winattr.width * bitsperpixel / 8;    /* Gerd Knorr's xawtv */
+	bitsperpixel = (depth+7) & 0xf8;
+	bytesperline *= bitsperpixel/8;
 
 	vfb.base = baseaddr;
 	vfb.height = Winattr.height;
@@ -1346,6 +1462,7 @@ GetFrameBuffer(void)
 	if (ioctl(tfd, VIDIOCSFBUF, &vfb) < 0) {
 		perror("ioctl VIDIOCSFBUF");
 	}
+#endif
 }
 
 
@@ -1390,11 +1507,11 @@ DoFullScreen(void)
 	back_pix = WhitePixel(display, DefaultScreen(display));
 	fore_pix = BlackPixel(display, DefaultScreen(display));
 
-	fmwin = XCreateWindow(display, DefaultRootWindow(display), fmsizehints.x, 
-					fmsizehints.y, fmsizehints.width, fmsizehints.height, 
-					borderwidth, CopyFromParent, InputOutput, 
+	fmwin = XCreateWindow(display, DefaultRootWindow(display), fmsizehints.x,
+					fmsizehints.y, fmsizehints.width, fmsizehints.height,
+					borderwidth, CopyFromParent, InputOutput,
 					CopyFromParent, valuemask, &fmWinattr);
-	
+
 
 	XSetWMNormalHints(display, fmwin, &fmsizehints);
 
@@ -1405,9 +1522,9 @@ DoFullScreen(void)
 
 	fmGC = XCreateGC(display, DefaultRootWindow(display), gcm, &gcv);
 
-	XSelectInput(display, fmwin, ButtonPressMask | ExposureMask | 
-					ButtonReleaseMask | PointerMotionMask | 
-					StructureNotifyMask | VisibilityChangeMask | KeyPressMask ); 
+	XSelectInput(display, fmwin, ButtonPressMask | ExposureMask |
+					ButtonReleaseMask | PointerMotionMask |
+					StructureNotifyMask | VisibilityChangeMask | KeyPressMask );
 
 	fmxwmhints.flags = StateHint;
 	fmxwmhints.initial_state = NormalState;
@@ -1432,9 +1549,9 @@ DoFullScreen(void)
 	}
 
 	usleep(50000L);
-	if (!XTranslateCoordinates(display, fmwin, DefaultRootWindow(display), 
-					-fmwinattr.border_width, 
-					-fmwinattr.border_width, 
+	if (!XTranslateCoordinates(display, fmwin, DefaultRootWindow(display),
+					-fmwinattr.border_width,
+					-fmwinattr.border_width,
 					&rx, &ry, &junkwin)) {
 		fprintf(stderr, "wmtv: error translating coordinates for fmwin");
 	}
@@ -1448,7 +1565,7 @@ DoFullScreen(void)
 	if (ioctl(tfd, VIDIOCSWIN, &vswin) < 0) {
 		perror("ioctl VIDIOCSWIN");
 	}
-	
+
 	ccapt = 1;
 	if (ioctl(tfd, VIDIOCCAPTURE, &ccapt) < 0) {
 		perror("ioctl VIDIOCCAPTURE");
@@ -1460,7 +1577,7 @@ DoFullScreen(void)
 		XF86VidModeGetAllModeLines(display, XDefaultScreen(display), &tml, &modelines);
 
 		for (i=0; i < tml; i++) {
-			if ((modelines[i]->hdisplay == fswidth) && 
+			if ((modelines[i]->hdisplay == fswidth) &&
 						(modelines[i]->vdisplay == fsheight)) {
 				fullscreenmode = modelines[i];
 				mode_present = TRUE;
@@ -1475,9 +1592,9 @@ DoFullScreen(void)
 		XRaiseWindow(display, fmwin);
 		if (mode_present) {
 
-			XF86VidModeSwitchToMode(display, XDefaultScreen(display), fullscreenmode); 
+			XF86VidModeSwitchToMode(display, XDefaultScreen(display), fullscreenmode);
 			XF86VidModeSetViewPort(display, XDefaultScreen(display), vswin.x, vswin.y);
-			XGrabPointer(display, fmwin, True, 0, GrabModeAsync, GrabModeAsync, 
+			XGrabPointer(display, fmwin, True, 0, GrabModeAsync, GrabModeAsync,
 						fmwin, None, CurrentTime);
 		}
 	}
@@ -1498,7 +1615,7 @@ RetScreen()
 
 	if (mode_present) {
 		for (i = 0; i < tml; i++) {
-			if ((modelines[i]->hdisplay == Winattr.width) && 
+			if ((modelines[i]->hdisplay == Winattr.width) &&
 							(modelines[i]->vdisplay==Winattr.height))
 				scm = modelines[i];
 		}
@@ -1517,7 +1634,7 @@ RetScreen()
 		scm->privsize = scmode.privsize;
 		scm->private = scmode.private;
 		*/
-	
+
 		XClearWindow(display, fmwin);
 		XFreeGC(display, fmGC);
 		XMapRaised(display, fmwin);
@@ -1562,6 +1679,7 @@ Usage(void)
 	fprintf(stderr, "\n");
 	fprintf(stderr, "wmtv v%s, Copyright (c) 1999 Wee Liang <wliang@tartarus.uwa.edu.au>\n", VERSION);
 	fprintf(stderr, "usage: wmtv [%s]\n", OPTIONS);
+	fprintf(stderr, "  -c, --device <file>\tsets video4linux device to use\n");
 	fprintf(stderr, "  -d, --display <host:vs>\tsets display name\n");
 	fprintf(stderr, "  -g, --geometry <{+-}XP{+-}YP>\tsets geometry\n");
 	fprintf(stderr, "  -b, --bpp\t\t\tdisplay color depth\n");
